@@ -189,7 +189,7 @@ inline int hal_ni_absDiffScalar8u8u  (const uchar* src_data, size_t src_step, uc
 Bitwise AND: _dst[i] = src1[i] & src2[i]_ @n
 Bitwise OR: _dst[i] = src1[i] | src2[i]_ @n
 Bitwise XOR: _dst[i] = src1[i] ^ src2[i]_ @n
-Bitwise NOT: _dst[i] = !src[i]_
+Bitwise NOT: _dst[i] = ~src[i]_
 @param src1_data first source image data
 @param src1_step first source image step
 @param src2_data second source image data
@@ -263,7 +263,7 @@ For 8s input type 128 is added to LUT index
 Destination should have the same element type and number of channels as lookup table elements
 @param src_data Source image data
 @param src_step Source image step
-@param src_type Sorce image type
+@param src_type Source image type
 @param lut_data Pointer to lookup table
 @param lut_channel_size Size of each channel in bytes
 @param lut_channels Number of channels in lookup table
@@ -307,9 +307,69 @@ Hamming distance between two vectors
 inline int hal_ni_normHammingDiff8u(const uchar* a, const uchar* b, int n, int cellSize, int* result) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
 //! @}
 
+/**
+@brief Generic norm of an array.
+@param src Source image
+@param src_step Source image
+@param mask Specified array region.
+@param mask_step Mask array step.
+@param width Source image dimensions
+@param height Source image dimensions
+@param type Element type of source image
+@param norm_type Type of the norm
+@param result Pointer to result output
+*/
+//! @addtogroup core_hal_interface_norm Absolute norm
+//! @{
+inline int hal_ni_norm(const uchar* src, size_t src_step, const uchar* mask, size_t mask_step, int width,
+                       int height, int type, int norm_type, double* result) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
+//! @}
+
+/**
+@brief Generic norm between two arrays.
+@param src1 First source image
+@param src1_step First source image
+@param src2 Second source image
+@param src2_step Second source image
+@param mask Specified array region.
+@param mask_step Mask array step.
+@param width Source image dimensions
+@param height Source image dimensions
+@param type Element type of source image
+@param norm_type Type of the norm
+@param result Pointer to result output
+*/
+//! @addtogroup core_hal_interface_norm Absolute norm
+//! @{
+inline int hal_ni_normDiff(const uchar* src1, size_t src1_step, const uchar* src2, size_t src2_step, const uchar* mask,
+                           size_t mask_step, int width, int height, int type, int norm_type, double* result) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
+//! @}
+
+/**
+@brief Convert array to another with specified type.
+@param src Source image
+@param src_step Source image
+@param dst Destination image
+@param dst_step Destination image
+@param width Source image dimensions
+@param height Source image dimensions
+@param sdepth Depth of source image
+@param ddepth Depth of destination image
+@param alpha Scale value
+@param beta Shift value
+*/
+//! @addtogroup core_hal_interface_convert Array convert
+//! @{
+inline int hal_ni_convertScale(const uchar* src, size_t src_step, uchar* dst, size_t dst_step, int width, int height,
+                               int sdepth, int ddepth, double alpha, double beta) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
+//! @}
+
 //! @cond IGNORED
 #define cv_hal_normHamming8u hal_ni_normHamming8u
 #define cv_hal_normHammingDiff8u hal_ni_normHammingDiff8u
+#define cv_hal_norm hal_ni_norm
+#define cv_hal_normDiff hal_ni_normDiff
+#define cv_hal_convertScale hal_ni_convertScale
 //! @endcond
 
 /**
@@ -394,7 +454,7 @@ inline int hal_ni_div64f(const double *src1_data, size_t src1_step, const double
 //! @}
 
 /**
-Computes reciprocial: _dst[i] = scale / src[i]_
+Computes reciprocal: _dst[i] = scale / src[i]_
 @param src_data source image data
 @param src_step source image step
 @param dst_data destination image data
@@ -403,7 +463,7 @@ Computes reciprocial: _dst[i] = scale / src[i]_
 @param height height of the images
 @param scale additional multiplier
  */
-//! @addtogroup core_hal_interface_reciprocial Element-wise reciprocial
+//! @addtogroup core_hal_interface_reciprocal Element-wise reciprocal
 //! @{
 inline int hal_ni_recip8u(const uchar *src_data, size_t src_step, uchar *dst_data, size_t dst_step, int width, int height, double scale) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
 inline int hal_ni_recip8s(const schar *src_data, size_t src_step, schar *dst_data, size_t dst_step, int width, int height, double scale) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
@@ -696,10 +756,28 @@ inline int hal_ni_dft1D(cvhalDFT *context, const uchar *src, uchar *dst) { retur
  */
 inline int hal_ni_dftFree1D(cvhalDFT *context) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
 
+/**
+@param src source data
+@param dst destination data
+@param depth depth of source
+@param nf OcvDftOptions data
+@param factors OcvDftOptions data
+@param scale OcvDftOptions data
+@param itab OcvDftOptions data
+@param wave OcvDftOptions data
+@param tab_size OcvDftOptions data
+@param n OcvDftOptions data
+@param isInverse OcvDftOptions data
+@param noPermute OcvDftOptions data
+ */
+inline int hal_ni_dft(const uchar* src, uchar* dst, int depth, int nf, int *factors, double scale, int* itab, void* wave,
+                         int tab_size, int n, bool isInverse, bool noPermute) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
+
 //! @cond IGNORED
 #define cv_hal_dftInit1D hal_ni_dftInit1D
 #define cv_hal_dft1D hal_ni_dft1D
 #define cv_hal_dftFree1D hal_ni_dftFree1D
+#define cv_hal_dft hal_ni_dft
 //! @endcond
 
 /**
@@ -764,14 +842,14 @@ inline int hal_ni_dctFree2D(cvhalDFT *context) { return CV_HAL_ERROR_NOT_IMPLEME
 Performs \f$LU\f$ decomposition of square matrix \f$A=P*L*U\f$ (where \f$P\f$ is permutation matrix) and solves matrix equation \f$A*X=B\f$.
 Function returns the \f$sign\f$ of permutation \f$P\f$ via parameter info.
 @param src1 pointer to input matrix \f$A\f$ stored in row major order. After finish of work src1 contains at least \f$U\f$ part of \f$LU\f$
-decomposition which is appropriate for determainant calculation: \f$det(A)=sign*\prod_{j=1}^{M}a_{jj}\f$.
+decomposition which is appropriate for determinant calculation: \f$det(A)=sign*\prod_{j=1}^{M}a_{jj}\f$.
 @param src1_step number of bytes between two consequent rows of matrix \f$A\f$.
 @param m size of square matrix \f$A\f$.
 @param src2 pointer to \f$M\times N\f$ matrix \f$B\f$ which is the right-hand side of system \f$A*X=B\f$. \f$B\f$ stored in row major order.
 If src2 is null pointer only \f$LU\f$ decomposition will be performed. After finish of work src2 contains solution \f$X\f$ of system \f$A*X=B\f$.
 @param src2_step number of bytes between two consequent rows of matrix \f$B\f$.
 @param n number of right-hand vectors in \f$M\times N\f$ matrix \f$B\f$.
-@param info indicates success of decomposition. If *info is equals to zero decomposition failed, othervise *info is equals to \f$sign\f$.
+@param info indicates success of decomposition. If *info is equals to zero decomposition failed, otherwise *info is equals to \f$sign\f$.
  */
 //! @addtogroup core_hal_interface_decomp_lu LU matrix decomposition
 //! @{
@@ -911,8 +989,26 @@ inline int hal_ni_gemm64fc(const double* src1, size_t src1_step, const double* s
 inline int hal_ni_minMaxIdx(const uchar* src_data, size_t src_step, int width, int height, int depth, double* minVal, double* maxVal,
                             int* minIdx, int* maxIdx, uchar* mask) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
 
+/**
+   @brief Finds the global minimum and maximum in an array.
+   @param src_data Source image
+   @param src_step Source image
+   @param width Source image dimensions
+   @param height Source image dimensions
+   @param depth Depth of source image
+   @param minVal Pointer to the returned global minimum and maximum in an array.
+   @param maxVal Pointer to the returned global minimum and maximum in an array.
+   @param minIdx Pointer to the returned minimum and maximum location.
+   @param maxIdx Pointer to the returned minimum and maximum location.
+   @param mask Specified array region.
+   @param mask_step Mask array step.
+*/
+inline int hal_ni_minMaxIdxMaskStep(const uchar* src_data, size_t src_step, int width, int height, int depth, double* minVal, double* maxVal,
+                                    int* minIdx, int* maxIdx, uchar* mask, size_t mask_step) { return CV_HAL_ERROR_NOT_IMPLEMENTED; }
+
 //! @cond IGNORED
 #define cv_hal_minMaxIdx hal_ni_minMaxIdx
+#define cv_hal_minMaxIdxMaskStep hal_ni_minMaxIdxMaskStep
 //! @endcond
 
 /**
@@ -1011,8 +1107,38 @@ inline int hal_ni_transpose2d(const uchar* src_data, size_t src_step, uchar* dst
 #define cv_hal_transpose2d hal_ni_transpose2d
 //! @endcond
 
-//! @}
+/**
+    @brief copyTo with mask
+    @param src_data, src_step Source image
+    @param dst_data, dst_step Destination image
+    @param width, height Image dimensions of source, destination and mask
+    @param type Type of source and destination images, for example CV_8UC1 or CV_32FC3
+    @param mask_data, mask_step, mask_type Mask
+*/
+inline int hal_ni_copyToMasked(const uchar* src_data, size_t src_step, uchar* dst_data, size_t dst_step, int width, int height,
+                             int type, const uchar* mask_data, size_t mask_step, int mask_type)
+{ return CV_HAL_ERROR_NOT_IMPLEMENTED; }
 
+//! @cond IGNORED
+#define cv_hal_copyToMasked hal_ni_copyToMasked
+//! @endcond
+
+/**
+ @ brief sum
+ @param src_data Source image data
+ @param src_step Source image step
+ @param src_type Source image type
+ @param width, height Source image dimensions
+ @param result Pointer to save the sum result to.
+ */
+inline int hal_ni_sum(const uchar *src_data, size_t src_step, int src_type, int width, int height, double *result)
+{ return CV_HAL_ERROR_NOT_IMPLEMENTED; }
+
+//! @cond IGNORED
+#define cv_hal_sum hal_ni_sum
+//! @endcond
+
+//! @}
 
 #if defined(__clang__)
 #pragma clang diagnostic pop
@@ -1026,26 +1152,23 @@ inline int hal_ni_transpose2d(const uchar* src_data, size_t src_step, uchar* dst
 #include "custom_hal.hpp"
 
 //! @cond IGNORED
-#define CALL_HAL_RET(name, fun, retval, ...) \
+
+#define CALL_HAL_RET2(name, fun, retval, ...) \
 { \
-    int res = __CV_EXPAND(fun(__VA_ARGS__, &retval)); \
+    int res = __CV_EXPAND(fun(__VA_ARGS__)); \
     if (res == CV_HAL_ERROR_OK) \
         return retval; \
     else if (res != CV_HAL_ERROR_NOT_IMPLEMENTED) \
         CV_Error_(cv::Error::StsInternal, \
-            ("HAL implementation " CVAUX_STR(name) " ==> " CVAUX_STR(fun) " returned %d (0x%08x)", res, res)); \
+        ("HAL implementation " CVAUX_STR(name) " ==> " CVAUX_STR(fun) " returned %d (0x%08x)", res, res)); \
 }
 
+#define CALL_HAL_RET(name, fun, retval, ...) \
+CALL_HAL_RET2(name, fun, retval, __VA_ARGS__, &retval)
 
 #define CALL_HAL(name, fun, ...) \
-{ \
-    int res = __CV_EXPAND(fun(__VA_ARGS__)); \
-    if (res == CV_HAL_ERROR_OK) \
-        return; \
-    else if (res != CV_HAL_ERROR_NOT_IMPLEMENTED) \
-        CV_Error_(cv::Error::StsInternal, \
-            ("HAL implementation " CVAUX_STR(name) " ==> " CVAUX_STR(fun) " returned %d (0x%08x)", res, res)); \
-}
+CALL_HAL_RET2(name, fun, ,__VA_ARGS__)
+
 //! @endcond
 
 #endif
